@@ -25,7 +25,7 @@ import traceback
 from functools import partial
 from contextlib import contextmanager
 
-from coconut.myparsing import (
+from coconut._pyparsing import (
     replaceWith,
     ZeroOrMore,
     Optional,
@@ -50,6 +50,7 @@ from coconut.constants import (
     use_computation_graph,
     py2_vers,
     py3_vers,
+    tabideal,
 )
 from coconut.exceptions import (
     CoconutException,
@@ -62,40 +63,63 @@ from coconut.exceptions import (
 # -----------------------------------------------------------------------------------------------------------------------
 
 
+def find_new_value(value, toklist, new_toklist):
+    """Find the value in new_toklist that corresponds to the given value in toklist."""
+    # find ParseResults by looking up their tokens
+    if isinstance(value, ParseResults):
+        if value._ParseResults__toklist == toklist:
+            new_value_toklist = new_toklist
+        else:
+            new_value_toklist = []
+            for inner_value in value._ParseResults__toklist:
+                new_value_toklist.append(find_new_value(inner_value, toklist, new_toklist))
+        return ParseResults(new_value_toklist)
+
+    # find other objects by looking them up directly
+    try:
+        return new_toklist[toklist.index(value)]
+    except ValueError:
+        complain(
+            lambda: CoconutInternalException(
+                "inefficient reevaluation of tokens: {} not in {}".format(
+                    value,
+                    toklist,
+                ),
+            ),
+        )
+        return evaluate_tokens(value)
+
+
 def evaluate_tokens(tokens):
     """Evaluate the given tokens in the computation graph."""
     if isinstance(tokens, str):
         return tokens
+
     elif isinstance(tokens, ParseResults):
+
         # evaluate the list portion of the ParseResults
         toklist, name, asList, modal = tokens.__getnewargs__()
         new_toklist = [evaluate_tokens(toks) for toks in toklist]
         new_tokens = ParseResults(new_toklist, name, asList, modal)
+
         # evaluate the dictionary portion of the ParseResults
         new_tokdict = {}
         for name, occurrences in tokens._ParseResults__tokdict.items():
             new_occurences = []
             for value, position in occurrences:
-                if isinstance(value, ParseResults) and value._ParseResults__toklist == toklist:
-                    new_value = new_tokens
-                else:
-                    try:
-                        new_value = new_toklist[toklist.index(value)]
-                    except ValueError:
-                        complain(lambda: CoconutInternalException("inefficient reevaluation of tokens: {} not in {}".format(
-                            value,
-                            toklist,
-                        )))
-                        new_value = evaluate_tokens(value)
+                new_value = find_new_value(value, toklist, new_toklist)
                 new_occurences.append(_ParseResultsWithOffset(new_value, position))
             new_tokdict[name] = occurrences
         new_tokens._ParseResults__accumNames.update(tokens._ParseResults__accumNames)
         new_tokens._ParseResults__tokdict.update(new_tokdict)
         return new_tokens
+
     elif isinstance(tokens, ComputationNode):
         return tokens.evaluate()
+
     elif isinstance(tokens, (list, tuple)):
         return [evaluate_tokens(inner_toks) for inner_toks in tokens]
+
     else:
         raise CoconutInternalException("invalid computation graph tokens", tokens)
 
@@ -568,3 +592,43 @@ def disable_outside(item, *elems):
     """
     for wrapped in disable_inside(item, *elems, **{"_invert": True}):
         yield wrapped
+
+
+def interleaved_join(outer_list, inner_list):
+    """Interleaves two lists of strings and joins the result.
+
+    Example: interleaved_join(['1', '3'], ['2']) == '123'
+    The first list must be 1 longer than the second list.
+    """
+    internal_assert(len(outer_list) == len(inner_list) + 1, "invalid list lengths to interleaved_join", (outer_list, inner_list))
+    interleaved = []
+    for xx in zip(outer_list, inner_list):
+        interleaved.extend(xx)
+    interleaved.append(outer_list[-1])
+    return "".join(interleaved)
+
+
+def handle_indentation(inputstr, add_newline=False):
+    """Replace tabideal indentation with openindent and closeindent."""
+    out_lines = []
+    prev_ind = None
+    for line in inputstr.splitlines():
+        new_ind_str, _ = split_leading_indent(line)
+        internal_assert(new_ind_str.strip(" ") == "", "invalid indentation characters for handle_indentation", new_ind_str)
+        internal_assert(len(new_ind_str) % tabideal == 0, "invalid indentation level for handle_indentation", len(new_ind_str))
+        new_ind = len(new_ind_str) // tabideal
+        if prev_ind is None:  # first line
+            indent = ""
+        elif new_ind > prev_ind:  # indent
+            indent = openindent * (new_ind - prev_ind)
+        elif new_ind < prev_ind:  # dedent
+            indent = closeindent * (prev_ind - new_ind)
+        else:
+            indent = ""
+        out_lines.append(indent + line)
+        prev_ind = new_ind
+    if add_newline:
+        out_lines.append("")
+    if prev_ind > 0:
+        out_lines[-1] += closeindent * prev_ind
+    return "\n".join(out_lines)
